use std::{fs::File, io::Read};
use env_logger::Builder;
use log::{error, info, LevelFilter};
use rkyv::{deserialize, rancor::Error, Archive, Deserialize, Serialize};

#[derive(Archive, Deserialize, Serialize, Debug, PartialEq)]
#[rkyv(compare(PartialEq), derive(Debug))]
struct ChunkHeader {
    chunk_type: [u8; 4],
    end_of_chunk: u32,
    children_count: u32
}

#[derive(Archive, Deserialize, Serialize, Debug, PartialEq)]
#[rkyv(compare(PartialEq), derive(Debug))]
struct Database {
    unknown: u32,
    version: u32,
    children_count: u32,
    id: u64,
    unknown1: [u8; 32],
    language: u16,
    persistent_id: u64,
    hash: [u8; 20]
}

#[derive(Archive, Deserialize, Serialize, Debug, PartialEq)]
#[rkyv(compare(PartialEq), derive(Debug))]
struct DataSet {
    data_type: u32
}

#[derive(Archive, Deserialize, Serialize, Debug, PartialEq)]
#[rkyv(compare(PartialEq), derive(Debug))]
struct AlbumItem {
    number_of_strings: u32,
    unknown: u16,
    album_id_for_track: u16,
    timestamp: u64,
    unknown1: u32
}

#[derive(Archive, Deserialize, Serialize, Debug, PartialEq)]
#[rkyv(compare(PartialEq), derive(Debug))]
struct TrackItem {
	number_of_strings: u32, // number of mhod's count
    unique_id: u32,
    visible: u32,
    filetype: u32,
    type1: u8,
    type2: u8,
    compilation_flag: u8,
    stars: u8,
    last_modified_time: u32,
    size: u32,
    length: u32,
    track_number: u32,
    total_tracks: u32,
    year: u32,
    bitrate: u32,
    sample_rate: u32,
    volume: u32,
    start_time: u32,
    stop_time: u32,
    soundcheck: u32,
    play_count: u32,
    play_count2: u32,
    last_played_time: u32,
    disc_number: u32,
    total_discs: u32,
    userid: u32,
    date_added: u32,
    bookmark_time: u32,
    dbid: u64,
    checked: u8,
    application_rating: u8,
    bpm: u16,
    artwork_count: u16,
    unk9: u16,
    artwork_size: u32,
    unk11: u32,
    sample_rate2: u32,
    date_released: u32,
    unk14: u32,
    unk15: u32,
    unk16: u32,
    skip_count: u32,
    last_skipped: u32,
    has_artwork: u8,
    skip_when_shuffling: u8,
    remember_playback_position: u8,
    flag4: u8,
    dbid2: u64,
    lyrics_flag: u8,
    movie_file_flag: u8,
    played_mark: u8,
    unk17: u8,
    unk21: u32,
    pregap: u32,
    sample_count: u64,
    unk25: u32,
    postgap: u32,
    unk27: u32,
    media_type: u32,
    season_number: u32,
    episode_number: u32,
    unk31: [u8; 28],
    gapless_data: u32,
    unk38: u32,
    gapless_track_flag: u16,
    gapless_album_flag: u16,
    unk39_hash: [u8; 20],
    unk40: [u8; 18],
    album_id: u16,
    mhii_link: u32
}

#[derive(Archive, Deserialize, Serialize, Debug, PartialEq)]
#[rkyv(compare(PartialEq), derive(Debug))]
struct Entry { // mhod
    entry_type: u32,
    unk1: u32,
    unk2: u32,
    position: u32,
    length: u32,
    unknown: u32,
    unk4: u32
}

enum ChunkState {
    Header,
    Data
}

fn db(data: &[u8]) {
    let mut state = ChunkState::Header;
    let mut chunk_header = None;
    let mut i = 0;
    while i < data.len() {
        state = match state {
            ChunkState::Header => {
                chunk_header = Some(rkyv::access::<ArchivedChunkHeader, Error>(&data[i..i+12]).unwrap());
                //info!("header: {:?}", chunk_header);
                i += 12;
                ChunkState::Data
            },
            ChunkState::Data => {
                let header = chunk_header.unwrap();
                let mut u = 0;
                // usize::try_from(header.end_of_chunk).unwrap() - 12
                match header.chunk_type {
                    [109, 104, 98, 100] => { //mhbd
                        u = usize::try_from(header.end_of_chunk).unwrap() - 12;
                        info!("val: {:?}", rkyv::access::<ArchivedDatabase, Error>(&data[i..i+u]).unwrap());
                    },
                    [109, 104, 115, 100] => { //mhsd
                        u = usize::try_from(header.end_of_chunk).unwrap() - 12;
                        info!("val: {:?}", rkyv::access::<ArchivedDataSet, Error>(&data[i..i+u]).unwrap());
                    },
                    [109, 104, 105, 97] => { // mhla
                        info!("AlbumList");
                        u = usize::try_from(header.end_of_chunk).unwrap() - 12;
                    },
                    [109, 104, 108, 97] => { // mhia
                        u = usize::try_from(header.end_of_chunk).unwrap() - 12;
                        info!("val: {:?}", rkyv::access::<ArchivedAlbumItem, Error>(&data[i..i+u]).unwrap());
                    },
                    [109, 104, 108, 116] => { // mhlt
                        info!("TrackList");
                        u = usize::try_from(header.end_of_chunk).unwrap() - 12;
                    },
                    [109, 104, 105, 116] => { // mhit
                        //info!("head: {}, {}", header.end_of_chunk, header.children_count);
                        u = usize::try_from(header.end_of_chunk).unwrap() - 12;
                        info!("val: {:?}", rkyv::access::<ArchivedTrackItem, Error>(&data[i..i+u]).unwrap());
                    },
                    [109, 104, 111, 100] => { // mhod
                        u = usize::try_from(header.children_count).unwrap() - 12;
                        //info!("head: {}, {}", header.end_of_chunk, header.children_count);
                        let header_offset: usize = (header.end_of_chunk + 4) as usize;
                        let str_end: usize = (header.children_count - 12) as usize;
                        let entry = rkyv::access::<ArchivedEntry, Error>(&data[i..]).unwrap();
                        info!("val: {:?}", &entry);
                        let mut bytes = Vec::new();

                        let mut h = i+header_offset;
                        while h < i+str_end {
                            if data[h] != 0 {
                                bytes.push(data[h]);
                            }
                            h+=1;
                        }
                        //info!("bytes: {:?}", &data[i+header_offset..i+str_end]);
                        let g = String::from_utf8(bytes).unwrap();
                        info!("str: {}", g);
                    }
                    _ => return
                }
                i += u;
                chunk_header = None;
                ChunkState::Header
            }
        }
    }
}

fn main() {

    // Initialize the logger with 'info' as the default level
    Builder::new()
        .filter(None, LevelFilter::Info)
        .init();

    let mut f = File::open("D:\\Documents\\iTunes\\iTunesDB").unwrap();
    let mut buf = Vec::new();
    match f.read_to_end(&mut buf) {
        Ok(n) => {
            let data = &buf[..n];
            db(data);
        },
        Err(e) => {
            error!("Error: {}",e);
        }
    }
}
